module Container.Initiality where

-- * Imports

-- ** Stdlib

open import Level
open import Function

open import Data.Unit
open import Data.Product

open import Relation.Binary.PropositionalEquality

-- ** Local

open import Pow
open import IProp

open import Container.Container
open import Container.Fixpoint
open import Container.Morphism

-- * Initiality of induction

initiality⊢ : {O : Set}{φ : O ▸ O}{o : O} 
               (Pred : {o : O} → μ φ o → Set)
               (x : μ φ o) →
               (f : {o : O}(x : μ φ o) → Pred x) → 
               (α : {o : O}(x : ⟦ φ ⟧ (μ φ) o) → □ φ (μ φ) Pred x → Pred < x >) →
               ({o : O}(xs : ⟦ φ ⟧ (μ φ) o) → ⊢ f < xs > ≡ α xs (f ∘ (proj₂ xs))) →
               ⊢ inductionCont φ Pred α x ≡ f x
initiality⊢ Pred < sh , pos > f α q = 
   pure trans ⊛ 
       (pure (cong (α (sh , pos))) ⊛
             (extensionality (λ p → initiality⊢ Pred (pos p) f α q))) ⊛ 
       (pure sym ⊛ q (sh , pos)) 

-- * Initiality of generalized induction

initiality'⊢ : {O : Set}{φ : O ▸ O}{o : O}(X : O → Set)
               (Pred : {o : O} → X o → Set)
               (x : μ φ o) →
               (α : {o : O} → ⟦ φ ⟧ X o → X o)
               (m : {o : O}(x : ⟦ φ ⟧ X o) → □ φ X Pred x → Pred (α x)) →
               (f : {o : O}(x : X o) → Pred x) → 
               ({o : O}(xs : ⟦ φ ⟧ X o) → ⊢ f (α xs) ≡ m xs (f ∘ proj₂ xs)) →
               ⊢ inductionCont' φ X Pred α m x ≡ (f ∘ foldCont α) x
initiality'⊢ X Pred x α m f q  = 
  initiality⊢ ((λ x → Pred (foldCont α x))) 
              x 
              (λ x → f (foldCont α x)) 
              (λ xs p → m ((proj₁ xs) , (λ p → foldCont α (proj₂ xs p))) p) 
              (λ xs → q ((proj₁ xs) , (λ p → foldCont α (proj₂ xs p))))


-- * Initiality of catamorphism

initialityCata⊢ : {O : Set}{o : O}{X : O → Set}{φ : O ▸ O} →
                  (α : ⟦ φ ⟧ X ⟶ X)
                  (x : μ φ o) → 
                  (f : μ φ ⟶ X) →
                  ({o : O}
                    (xs : ⟦ φ ⟧ (μ φ) o) → 
                    ⊢ f < xs > ≡ (α ∘ ⟦ φ ⟧Morph (\{o} → f {o})) xs) → 
                  ⊢ foldCont α x ≡ f x
initialityCata⊢ {O}{o}{X}{φ} α < sh , pos > f q = 
    pure trans ⊛ 
      (pure (cong (λ p → α (sh , p))) ⊛ 
          extensionality (λ p → initialityCata⊢ α (pos p) f q)) ⊛ 
      (pure sym ⊛ q (sh , pos))

{- Or:

= initiality⊢ (\{o} _ → X o)
                                                   x 
                                                   f
                                                   (\x p → α (proj₁ x , p))
                                                   q
-}

-- * Fusion law

fusionCata⊢ : {O : Set}{o : O}{X : O → Set}{Y : O → Set}{φ : O ▸ O} →
              (α : ⟦ φ ⟧ X ⟶ X)
              (x : μ φ o) → 
              (h : X ⟶ Y)
              (β : ⟦ φ ⟧ Y ⟶ Y) →
              ({o : O}
                 (xs : ⟦ φ ⟧ X o) → 
                    ⊢ h (α xs) ≡ β (⟦ φ ⟧Morph (\{o} → h {o}) xs)) → 
                  ⊢ foldCont β x ≡ h (foldCont α x)
fusionCata⊢ {φ = φ} α x h β eq = 
  initialityCata⊢ β x (h ∘ foldCont α)
                      (λ xs → eq (⟦ φ ⟧Morph (λ {x} → foldCont (λ {i} → α {i}) {x}) xs))

-- * Fusing natural transformations

-- TODO This is not actually used. Remove? 

module Fuse {O : Set}{O+ : Set}{o+ : O+}{X : O → Set}
            {φ : O ▸ O}{φ+ : O+ ▸ O+}{reO : O+ → O}
            (η : φ+ ⇒[ reO ] φ)
            (α : ⟦ φ ⟧ X ⟶ X) where

       unfused : μ φ+ o+ → X (reO o+)
       unfused x = foldCont α (foldCont (λ {o+} xs → <_> {O}{φ}{reO o+} (⟦ η ⟧⇒ (μ φ) o+ xs)) x)

       fused : μ φ+ o+ → X (reO o+)
       fused x = foldCont (λ {o+} xs → α (⟦ η ⟧⇒ X o+ xs)) x

       fuseNT⊢ : (t : μ φ+ o+) → ⊢ fused t ≡ unfused t
       fuseNT⊢ t = fusionCata⊢ (<_> {O}{φ}{reO _} ∘ ⟦ η ⟧⇒ (μ φ) _) t (foldCont α) (α ∘ ⟦ η ⟧⇒ X _) 
                               (λ xs → pure (cong (λ p → α (u η (proj₁ xs), p))) ⊛
                                            extensionality (λ x → pure (lemmaSubst (λ{i} → foldCont (λ{i} → α{i}) {i}) 
                                                                                   (proj₂ xs (f η x))
                                                                                   (sim⊢ η (proj₁ xs) x))))

-- * Outro

-- Local Variables:
-- mode: outline-minor
-- outline-regexp: "-- [*\f]+"
-- outline-level: outline-level
-- End:                      
