open import Pow

open import Container.Container
open import Container.Fixpoint
open import Container.Initiality
open import Container.Morphism

open import Ornament.Ornament
open import Ornament.CoalgebraicOrnament

open import Example.Container

open import Data.Empty
open import Data.Unit
open import Data.Bool
open import Data.Nat
open import Data.Product
open import Data.Sum

open import Relation.Binary.PropositionalEquality

module Example.CoalgebraicOrnament where

-- * Balanced tree

data TreeConst : Set where
  `leaf : TreeConst
  `node : TreeConst

TreeCont : ⊤ ▸ ⊤
TreeCont = record { S = \_ → TreeConst
              ; P = PTree
              ; n = nTree }
     where PTree : TreeConst → Set
           PTree `leaf = ⊥
           PTree `node = Bool

           nTree : {sh : TreeConst} → PTree sh → ⊤
           nTree p = tt

balancedTreeD : (⊤ × ℕ) ▸ (⊤ × ℕ)
balancedTreeD = coalgebraicOrnament TreeCont top down
  where top : ℕ → S TreeCont tt → ℕ
        top n `leaf = 0
        top n `node = suc n

        down : (n : ℕ){sh : S TreeCont tt} → (p : P TreeCont sh) → ℕ
        down n {`leaf} ()
        down n {`node} p = n



balancedTree : ℕ → Set
balancedTree n = μ balancedTreeD (tt , n)

leaf : balancedTree 0
leaf = < ((`leaf , (0 , refl)) , \p → ⊥-elim p) >

node : ∀{n} → balancedTree n → balancedTree n → balancedTree (suc n)
node {n} l r = < ((`node , (n , refl)) , \p → if p then l else r) >


-- * Fin

FinD : (⊤ × ℕ) ▸ (⊤ × ℕ)
FinD = coalgebraicOrnament NatCont top down
  where top : ℕ → S NatCont tt → ℕ
        top n `0 = suc n
        top n `suc = suc n

        down : (n : ℕ){sh : S NatCont tt} → (p : P NatCont sh) → ℕ
        down n {`0} ()
        down n {`suc} tt = n

Fin : ℕ → Set
Fin n = μ FinD (tt , n)

fz : {n : ℕ} → Fin (suc n)
fz {n} = < ((`0 , (n , refl)) , \p → ⊥-elim p) >

fsuc : {n : ℕ} → Fin n → Fin (suc n)
fsuc {n} x = < ((`suc , (n , refl)) , \p → x) >

-- * Red black tree?

data Color : Set where
  Red Black : Color

-- RBT is not an instance of top-down indexing: the indexing flows
-- from black node to black node, not from root to leaves. On a black
-- node, there is a non-deterministic choice of the color of the
-- descendent.

-- Hence, without pre-decorating the tree with the color of black
-- nodes, we cannot write the following coalgebraic ornament.

RBTD : (⊤ × ℕ × Color) ▸ (⊤ × ℕ × Color)
RBTD = coalgebraicOrnament TreeCont top {!!}
  where top : ℕ × Color → TreeConst → ℕ × Color
        top (n , c) `leaf = 0 , Black
        top (n , Red) `node = n , Red
        top (n , Black) `node = suc n , Black

        down : ℕ × Color → {sh : TreeConst}(p : P TreeCont sh) → ℕ × Color
        down (n , c) {`leaf} ()
        down (n , Red) {`node} p = n , Black
        down (n , Black) {`node} p = {!!} 


