module Example.Container where

-- * Imports 

-- ** Stdlib

open import Data.Empty
open import Data.Unit
open import Data.Sum
open import Data.Product
open import Data.Nat

-- ** Local

open import Container.Container
open import Container.Fixpoint

-- * Example: Nat

data NatConst : Set where
  `0 : NatConst
  `suc : NatConst

module NatDef where
  SNat : ⊤ → Set
  SNat _ = NatConst

  PNat : {o : ⊤} → SNat o → Set
  PNat `0 = ⊥
  PNat `suc = ⊤

  nNat : {o : ⊤}{sh : SNat o} → PNat sh → ⊤
  nNat _ = tt

NatCont : ⊤ ▸ ⊤
NatCont = record { S = NatDef.SNat 
                 ; P = NatDef.PNat
                 ; n = NatDef.nNat }

Nat : Set
Nat = μ NatCont tt

ze : Nat
ze = < (`0 , ⊥-elim) >

su : Nat → Nat
su n = < (`suc , \ _ → n) >

-- * Example: List

module ListCont' (A : Set) where

  SList : ⊤ → Set
  SList tt = ⊤ ⊎ A

  PList : SList tt → Set
  PList (inj₁ tt) = ⊥
  PList (inj₂ a) = ⊤

  nList : {sh : SList tt} → PList sh → ⊤
  nList _ = tt

ListCont' : Set → ⊤ ▸ ⊤
ListCont' A = record { S = ListCont'.SList A
                     ; P = ListCont'.PList A
                     ; n = ListCont'.nList A }

List : Set → Set
List A = μ (ListCont' A) tt

nil : {A : Set} → List A
nil = < (inj₁ tt , ⊥-elim) >

cons : {A : Set} → A → List A → List A
cons a xs = < (inj₂ a , λ _ → xs) >



-- * Example: Vector

module VecCont' (A : Set) where

  SVec : ℕ → Set
  SVec zero = ⊤
  SVec (suc n) = A

  PVec : {n : ℕ} → SVec n → Set
  PVec {zero} tt = ⊥
  PVec {suc n} a = ⊤

  nVec : {n : ℕ}{sh : SVec n} → PVec {n} sh → ℕ
  nVec {zero} ()
  nVec {suc n} tt = n

VecCont' : Set → ℕ ▸ ℕ
VecCont' A = record { S = VecCont'.SVec A
                    ; P = λ {n} → VecCont'.PVec A {n}
                    ; n = VecCont'.nVec A }

Vec : Set → ℕ → Set
Vec A n = μ (VecCont' A) n

vnil : {A : Set} → Vec A 0
vnil = < (tt , λ ()) >

vcons : {A : Set}{n : ℕ} → A → Vec A n → Vec A (suc n)
vcons a xs = < (a , λ _ → xs) >

-- * Outro

-- Local Variables:
-- mode: outline-minor
-- outline-regexp: "-- [*\f]+"
-- outline-level: outline-level
-- End:                      
