module Example.Ornament where 

open import Container.Container
open import Container.Fixpoint

open import Ornament.Ornament

open import Example.Container

open import Data.Empty
open import Data.Unit
open import Data.Product
open import Data.Sum
open import Data.Nat

open import Relation.Binary.PropositionalEquality

-- ** Example: List extends Nat

forgetList : ⊤ → ⊤
forgetList tt = tt

ListOrn : Set → Ornament1 NatCont ⊤ forgetList
ListOrn A = record { extend = aList
                   ; refine = \ e p → tt
                   ; coherence⊢ = \x e p → refl }
      where aList : ⊤ → NatConst → Set
            aList tt `0 = ⊤
            aList tt `suc = A

ListCont : Set → ⊤ ▸ ⊤
ListCont A = ⟨ ListOrn A ⟩o

List' : (A : Set) → Set
List' A = μ (ListCont A) tt

nil' : {A : Set} →  List' A
nil' = < ((`0 , tt) , ⊥-elim) >

cons' : {A : Set} → A → List' A → List' A
cons' a tl = < ((`suc , a) , \ _ → tl) >

-- ** Example: Vector refines List

forgetVector : ℕ → ⊤
forgetVector _ = tt

VectorOrn : (A : Set) → Ornament1 (ListCont' A) ℕ forgetVector
VectorOrn A = record { extend = extendVector
                     ; refine = \{o+}{sh} → refineVector {o+} {sh}
                     ; coherence⊢ = \{o+} → coherenceVector {o+} }
               where extendVector : (o' : ℕ) → ⊤ ⊎ A → Set
                     extendVector zero (inj₁ tt) = ⊤
                     extendVector zero (inj₂ x) = ⊥
                     extendVector (suc n) (inj₁ tt) = ⊥
                     extendVector (suc n) (inj₂ x) = ⊤
                     
                     refineVector : {o' : ℕ}{sh : ⊤ ⊎ A}(e : extendVector o' sh) →
                                     P (ListCont' A) sh → ℕ
                     refineVector {zero} {inj₁ tt} e ()
                     refineVector {zero} {inj₂ x} () p
                     refineVector {suc n} {inj₁ tt} () p
                     refineVector {suc n} {inj₂ x} e p = n

                     coherenceVector : {o' : ℕ}(sh : ⊤ ⊎ A)(e : extendVector o' sh)
                                       (p : P (ListCont' A) sh) → tt ≡ tt
                     coherenceVector sh e p = refl

VectorCont : Set → ℕ ▸ ℕ
VectorCont A = ⟨ VectorOrn A ⟩o

Vector : (n : ℕ)(A : Set) → Set
Vector n A = μ (VectorCont A) n

vnil' : {A : Set} → Vector 0 A
vnil' = < ((inj₁ tt , tt) , \p → ⊥-elim p ) >

vcons' : {n : ℕ}{A : Set} → A → Vector n A → Vector (1 + n) A
vcons' {n} {A} a xs = < (((inj₂ a) , tt) , \_ → xs) >
