module Lifting.Patch where

-- * Imports

-- ** Stdlib

open import Function

open import Relation.Binary.PropositionalEquality

open import Data.Product
open import Data.Unit

-- ** Local

open import Pow

open import Container.Container
open import Container.Fixpoint

open import Ornament.Ornament
open import Ornament.LiftOrnament
open import Ornament.OrnamentalAlgebra

open import Lifting.Universe

-- * Patch of type scheme

-- ** Definition

-- TODO: add a command to introduce a (polymorphic) parameter set

data Patch : Type → Set₁ where
  μ⁺[_]_[_∣_]⇒_ : ∀ {O : Set}{φ}{o : O}{T} → 
             {O+ : Set}(forget+ : O+ → O)(υ : Ornament1 φ O+ forget+)(o+ : O+)(q : forget+ o+ ≡ o)
             (T+ : Patch T) → 
             Patch (μ φ [ o ]⇒ T)
  μ⁺[_]_[_∣_]×_ : ∀ {O : Set}{φ}{o : O}{T} → 
            {O+ : Set}(forget+ : O+ → O)(υ : Ornament1 φ O+ forget+)(o+ : O+)(q : o ≡ forget+ o+) 
            (T+ : Patch T) →
            Patch (μ φ [ o ]× T)
  `⊤ : Patch `⊤ 

-- ** Interpretation

-- TODO: interpret PatchType as a Type once Type is type-dependent enough

⟦_⟧PatchType : ∀{T} → Patch T → Set
⟦ μ⁺[ forget+ ] υ [ o+ ∣ q ]⇒ T+ ⟧PatchType = (t+ : μ ⟨ υ ⟩o o+) → ⟦ T+ ⟧PatchType
⟦ μ⁺[ forget+ ] υ [ o+ ∣ q ]× T+ ⟧PatchType = Σ[ t+ ∶ μ ⟨ υ ⟩o o+ ] ⟦ T+ ⟧PatchType
⟦ `⊤ ⟧PatchType = ⊤


-- * Lifter object

Lifter : (T : Type) → Patch T → ⟦ T ⟧LT → Set
Lifter (μ φ [ .(forget+ o+) ]⇒ T) (μ⁺[ forget+ ] υ [ o+ ∣ refl ]⇒ T+) f = 
    (x : μ φ (forget+ o+)) → μ (liftCont υ) (o+ , x) → Lifter T T+ (f x)
Lifter (μ φ [ .(forget+ o+) ]× T) (μ⁺[ forget+ ] υ [ o+ ∣ refl ]× T+) (a , b) = 
    μ (liftCont υ) (o+ , a) × Lifter T T+ b
Lifter `⊤ `⊤ tt = ⊤

-- * Lifted function

⟦_⟧PatchFun : ∀{T} → (P : Patch T)(f : ⟦ T ⟧LT) → Lifter T P f → ⟦ P ⟧PatchType
⟦ μ⁺[ forget+ ] υ [ o+ ∣ refl ]⇒ T+ ⟧PatchFun f f++ = 
  λ t+ → ⟦ T+ ⟧PatchFun (f (forgetOrn υ t+)) (f++ (forgetOrn υ t+) (doLift υ t+))
⟦ μ⁺[ forget+ ] υ [ o+ ∣ refl ]× T+ ⟧PatchFun (t , x) (t++ , x++) = 
  downLift υ t++ , ⟦ T+ ⟧PatchFun x x++
⟦ `⊤ ⟧PatchFun tt tt = tt

-- * Outro

-- Local Variables:
-- mode: outline-minor
-- outline-regexp: "-- [*\f]+"
-- outline-level: outline-level
-- End:                      
