-- * Imports

-- ** Stdlib

open import Function

open import Relation.Binary.PropositionalEquality

open import Data.Product

open import Pow

-- ** Local

open import Container.Container
open import Container.Fixpoint
open import Container.Initiality

open import Ornament.Ornament

module Ornament.CoalgebraicOrnament {O : Set}{X : O → Set}
                                    (φ : O ▸ O) 
                                    (top : {o : O} → X o → S φ o → X o)
                                    (down : {o : O} → (x : X o){sh : S φ o} → (p : P φ sh) → X (n φ p)) where

-- * The Algebraic Ornament

-- The (trivial) forget:
forget : Σ O X → O
forget = proj₁

-- Ornament φ with the algebra α:
coalgebraicOrn : Ornament1 φ (Σ O X) forget
coalgebraicOrn = record { extend = extendAlgOrn
                        ; refine = refineAlgOrn
                        ; coherence⊢ = coherenceAlgOrn }

         where extendAlgOrn : (o+ : Σ O X) → S φ (forget o+) → Set
               extendAlgOrn (o , x) sh = Σ[ x' ∶ X o ] top x' sh ≡ x
       
               refineAlgOrn : {o+ : Σ O X}{sh : S φ (forget o+)}
                              (e : extendAlgOrn o+ sh) → P φ sh → Σ O X
               refineAlgOrn (x , eq) p = n φ p , down x p

               coherenceAlgOrn : {o+ : Σ O X}(sh : S φ (forget o+))
                                 (e : extendAlgOrn o+ sh)
                                 (p : P φ sh) →
                                   proj₁ (refineAlgOrn e p) ≡ n φ p
               coherenceAlgOrn sh e p = refl

coalgebraicOrnament : Σ O X ▸ Σ O X
coalgebraicOrnament = ⟨ coalgebraicOrn ⟩o



-- * Outro

-- Local Variables:
-- mode: outline-minor
-- outline-regexp: "-- [*\f]+"
-- outline-level: outline-level
-- End:                      
