module Ornament.Ornament where

-- * Imports

-- ** Stdlib

open import Function

open import Relation.Binary.PropositionalEquality

open import Data.Product

-- ** Local

open import Container.Container
open import Container.CartesianMorphism

-- * Definition

record Ornament {I O : Set}                   
                (φ : I ▸ O)(I+ O+ : Set)
                -- We can refine indices:
                (forgetI+ : I+ → I)
                (forgetO+ : O+ → O) : Set1 where
  field
    -- We can give more precise shapes:
    extend : (o+ : O+)(sh : S φ (forgetO+ o+)) → Set

    -- We can be more restrictive for the next index:
    refine : {o+ : O+}{sh : S φ (forgetO+ o+)}(e : extend o+ sh) → P φ sh → I+
    coherence⊢ : {o+ : O+}(sh : S φ (forgetO+ o+))(e : extend o+ sh)(p : P φ sh) →
                    forgetI+ (refine e p) ≡ n φ p 

open Ornament public

-- * Restriction to one index

Ornament1 : {O : Set}(φ : O ▸ O)(O+ : Set)(forget+ : O+ → O) → Set1
Ornament1 φ O+ forget+ = Ornament φ O+ O+ forget+ forget+

-- * Extension to a container

⟨_⟩o : {I I+ O O+ : Set}{φ : I ▸ O}{forgetI+ : I+ → I}{forgetO+ : O+ → O}
       (υ : Ornament φ I+ O+ forgetI+ forgetO+) → I+ ▸ O+
⟨_⟩o {φ = φ}{forgetO+ = forgetO+} υ = 
     record { S = \o+ → Σ[ sh ∶ S φ (forgetO+ o+) ] extend υ o+ sh
            ; P = P φ ∘ proj₁
            ; n = \{o+}{sh} → refine υ (proj₂ sh) }


-- * Ornament ≅ CartesianMorphism

toCartesian : {I I+ O O+ : Set}{φ : I ▸ O}{forgetI+ : I+ → I}{forgetO+ : O+ → O}
              (υ : Ornament φ I+ O+ forgetI+ forgetO+) → ⟨ υ ⟩o ⇒c[ forgetI+ ∣ forgetO+ ] φ
toCartesian υ =  
  record { forget = proj₁
         ; pull = refl
         ; simForget⊢ = \sh p → (coherence⊢ υ (proj₁ sh) (proj₂ sh) p) }


fromCartesian : {I+ I O+ O : Set}{φ+ : I+ ▸ O+}{φ : I ▸ O}{forgetI+ : I+ → I}{forgetO+ : O+ → O} → 
                φ+ ⇒c[ forgetI+ ∣ forgetO+ ] φ → Ornament φ I+ O+ forgetI+ forgetO+
fromCartesian {I+}{O+ = O+}{φ+ = φ+}{φ}{forgetI+}{forgetO+} τ = 
  record { extend = \o+ sh → Σ[ sh+ ∶ S φ+ o+ ] sh ≡ forget τ sh+ 
         ; refine = refineC 
         ; coherence⊢ = coherenceC }

      where refineC : {o+ : O+} {sh : S φ (forgetO+ o+)}
                     (e : Σ (S φ+ o+) (λ sh+ → sh ≡ forget τ sh+)) →
                     P φ sh → I+
            refineC (sh+ , q) p rewrite q = n φ+ (subst id (pull τ) p)
        
            coherenceC : {o+ : O+} (sh : S φ (forgetO+ o+))
                         (e : Σ (S φ+ o+) (λ sh+ → sh ≡ forget τ sh+)) (p : P φ sh) →
                         forgetI+ (refineC e p) ≡ n φ p
            coherenceC .(forget τ sh+) (sh+ , refl) p = simForget⊢ τ sh+ p

-- * Outro

-- Local Variables:
-- mode: outline-minor
-- outline-regexp: "-- [*\f]+"
-- outline-level: outline-level
-- End:                      
