-- * Imports

-- ** Stdlib

open import Data.Empty
open import Data.Unit
open import Data.Nat
open import Data.Sum
open import Data.Product

open import Relation.Binary.PropositionalEquality

-- ** Local

open import Container.Container
open import Container.Morphism

open import Ornament.Ornament


module Ornament.PullbackOrnament {O O+ J+ : Set}
                                 (φ : O ▸ O)(forgetO+ : O+ → O)(forgetJ+ : J+ → O)
                                 (τ : Ornament1 φ O+ forgetO+)(τ' : Ornament1 φ J+ forgetJ+) where

-- * Library:

-- We need a few wacky substitutions.

-- TODO Can we get rid of that?

subst₂' : ∀{A : Set} {B : A → Set} (P : (a : A) → B a → Set)
          {x₁ x₂ y₁ y₂} → (q : x₁ ≡ x₂) → y₁ ≡ y₂ → P x₁ y₁ → P x₂ (subst B q y₂)
subst₂' P refl refl p = p

cong₃ : ∀{A : Set}{B : A → Set}{C : (a : A) → B a → Set}{D : Set}
        (f : (a : A)(b : B a) → C a b → D ) 
        {x y : A}{u v : B x}{s t : C x u} → 
        (qa : x ≡ y)(qb : u ≡ v) → s ≡ t → f x u s ≡ f y (subst B qa v) (subst₂' C qa qb t)
cong₃ f refl refl refl = refl

-- * Pullback ornament

O□ : Set 
O□ = Σ[ x ∶ O+ ] Σ[ y ∶ J+ ] Σ[ z ∶ O ] z ≡ forgetO+ x × z ≡ forgetJ+  y

forget□ : O□ → O
forget□ (x , y , z , _ , _) = z

τ□τ' : Ornament1 φ O□ forget□
τ□τ' = record { extend = extend□
              ; refine = \{o□}{sh□} e p → refine□ {o□} {sh□} e p
              ; coherence⊢ = \{o□} sh□ e p → coherence□⊢ {o□} sh□ e p  }
     where

           extend□ : (o□ : O□) → S φ (forget□ o□) → Set
           extend□ (x , y , z , qx , qy ) sh = extend τ x (subst (\z → S φ z) qx sh) 
                                             × extend τ' y (subst (\z → S φ z) qy sh)

           refine□ : {o□ : O□}{sh□ : S φ (forget□ o□)}(e : extend□ o□ sh□) → P φ sh□ → O□
           refine□ {(x , y , z , qx , qy)}{sh} (e1 , e2) p = 
             refine τ {x} {subst (\z → S φ z) qx sh} e1 (subst₂' (\z sh → P φ {z} sh) qx refl p) , 
             refine τ' {y} {subst (\z → S φ z) qy sh} e2 (subst₂' (\z sh → P φ {z} sh) qy refl p) , 
             n φ {z} {sh} p ,
             let q = coherence⊢ τ {x} (subst (\z → S φ z) qx sh) e1 (subst₂' (\z → P φ {z}) qx refl p) in 
               trans (cong₃ (\x y → n φ {x} {y}) qx refl refl) (sym q) , 
             let q = coherence⊢ τ' {y} (subst (S φ) qy sh) e2 (subst₂' (\z → P φ {z}) qy refl p) in 
               trans (cong₃ (\x y → n φ {x} {y}) qy refl refl) (sym q)

           coherence□⊢ : {o□ : O□}(sh□ : S φ (forget□ o□))(e : extend□ o□ sh□)(p : P φ sh□) →
                         n φ p ≡ forget□ (refine□ {o□} {sh□} e p)
           coherence□⊢ sh□ e p = refl


-- * Outro

-- Local Variables:
-- mode: outline-minor
-- outline-regexp: "-- [*\f]+"
-- outline-level: outline-level
-- End:                      
